/**** SimCore since 2002-05-15 by Kenji KISE ****/ 
/************************************************/
#include "define.h"

/************************************************/
void memory_system::ld_8byte(data_t *a, data_t *d){
  mm->ld_8byte(a, d);
}

void memory_system::st_8byte(data_t *a, data_t *d,
			     DATA_TYPE msk){
  mm->st_8byte(a, d, msk);
}

/************************************************/
memory_system::~memory_system(){}

/************************************************/
memory_system::memory_system(Env *ev_t){
  ev = ev_t;
  mm = new main_memory();
}

/************************************************/
void memory_system::ld_nbyte(int n, data_t *a, 
                             data_t *d){
  if(*a%n!=0) 
    printf("*** ld_nbyte %d miss-align.\n", n);
  if(n!=1 && n!=2 && n!=4 && n!=8)
    printf("*** Error in ld_nbyte %d.\n", n);
  
  ld_8byte(a, d); /** Type Conversion **/
  if(n==8) return;

  int offset = *a & 7;
  if(n==4){
    DATA_TYPE dt = (*d>>(offset*8))&MASK32;
    *d=dt;
  }
  else if(n==2){
    DATA_TYPE dt = (*d>>(offset*8))&MASK16;
    *d=dt;
  }
  else if(n==1){
    DATA_TYPE dt = (*d>>(offset*8))&MASK08;
    *d=dt;
  }
}

/************************************************/
void memory_system::st_nbyte(int n, data_t *a, 
                             data_t *d){
  if(*a%n!=0)
    printf("*** st_nbyte %d miss-alig.\n", n);
  if(n!=1 && n!=2 && n!=4 && n!=8)
    printf("*** Error in st_nbyte %d.\n", n);

  int offset = *a & 7;
  DATA_TYPE mask=0;

  if(n==4){
    mask = ~(MASK32 << offset*8);
    DATA_TYPE dt = (*d&MASK32)<<offset*8;
    *d=dt;
  }
  else if(n==2){
    mask = ~(MASK16 << offset*8);
    DATA_TYPE dt = (*d&MASK16)<<offset*8;
    *d=dt;
  }
  else if(n==1){
    mask = ~(MASK08 << offset*8);
    DATA_TYPE dt = (*d&MASK08)<<offset*8;
    *d=dt;
  }

  st_8byte(a, d, mask); /** Type Conversion **/
}

/** Implementation of raw 32-bit memory space  **/
/************************************************/
main_memory::main_memory(){
  for(int i=0; i<BLOCK_TABLE_SIZE; i++){
    block_table[i]=NULL;
  }
}

/************************************************/
int main_memory::load_all(data_t *a, data_t *d){
  static int i=0;
  static int j=0;
  while(i<BLOCK_TABLE_SIZE){
    if(block_table[i]!=NULL)
      for(;j<(BLOCK_SIZE/DATA_T_SIZE);j++){
	data_t *bt = block_table[i];
	if(bt[j]!=0){
	  *a = i<<BLOCK_MASK_BIT | (j<<3);
	  *d = bt[j];
	  j++;
	  return 1; /* return the data */
	}
      }
    i++;
    j=0;
  }
  return 0; /* all data is passed */
}

/************************************************/
data_t *main_memory::allocblock(data_t *a){
  int n = BLOCK_SIZE/DATA_T_SIZE;
  data_t *ret =new data_t[n]; /* allocate */
  for(int i=0; i<n; i++) ret[i]=0;

  uint_32t adr = *a;
  block_table[adr>>BLOCK_MASK_BIT] = ret;
  if(ret==NULL){
    printf("** Error in allocblock.\n");
    exit(0);
  }
  return ret;
}

/************************************************/
void main_memory::ld_8byte(data_t *a, data_t *d){
  uint_32t adr = *a;
  data_t *bt = block_table[adr>>BLOCK_MASK_BIT];
  data_t *ptr = (bt!=NULL) ? bt : allocblock(a);
  int offset = (adr & BLOCK_MASK)>>DATA_T_BIT;
  *d = ptr[offset];
}

/************************************************/
void main_memory::st_8byte(data_t *a, data_t *d, 
			   DATA_TYPE msk){
  uint_32t adr = *a;
  data_t *bt = block_table[adr>>BLOCK_MASK_BIT];
  data_t *ptr = (bt!=NULL) ? bt : allocblock(a);
  int offset = (adr & BLOCK_MASK)>>DATA_T_BIT;
  ptr[offset] = (msk==0) ? *d :
    (ptr[offset] & msk) | *d;
}
/************************************************/
